""" Provide patched-up versions of common Python DOM implementations. PXTL
    requires parts of DOM Level 3 Core and Load which are not yet implemented
    in current versions of these implementations, and it also gets tripped up
    by certain bugs in their DOM Level 2 Core code. This module patches them
    so they work better.
"""

__all__= [
  'getImplementation'
]

import urllib, urlparse
from pxtl.constants import *

_implementations= {}

def getImplementation(name):
  """ Return the DOMImplementation object for a fixed version of known
      implementations 'minidom' and '4DOM'. Cache newly-created
      implementations to try to avoid creating multiple implementation
      objects.
  """
  if _implementations.has_key(name):
    return _implementations[name]
  if name.lower()=='4dom':
    implementation= FourDOMImplementation()
  elif name.lower()=='minidom':
    implementation= MinidomImplementation()
  else:
    raise KeyError('Implementation %s not known. Try 4DOM or minidom.' % name)
  _implementations[name]= implementation
  return implementation


class FakeDOMConfiguration:
  """ A object with DOMConfiguration's interface, that does nothing.
  """
  def setParameter(self, name, value):
    pass
  def getParameter(self, name):
    return None
fakeDOMConfiguration= FakeDOMConfiguration()


class L2Implementation:
  """ An object delegating to a Level 2 DOMImplementation that implements the
      DOM Level 3 features XML Base, DOMParser and namespace on top of it.
      Subclasses must provide a getDocument function that returns a DOM tree
      for a given URI.

      Their constructor must also set 'implementation' to point to their
      implementation's DOMImplementation object, and 'node' and 'document' to
      point to their classes implementing those interfaces. These classes will
      be patched if necessary to implement the new features.
  """
  implementation= None
  config= fakeDOMConfiguration
  [MODE_SYNCHRONOUS, MODE_ASYNCHRONOUS]= range(1, 3)
  def __init__(self):
    if not self.implementation.__dict__.has_key('__fixdom_patched'):
      self.implementation.__dict__['__fixdom_patched']= True
      self.patchModule()
  def hasFeature(self, feature, version):
    return self.implementation.hasFeature(feature, version)
  def getFeature(self, feature, version):
    return self.implementation.getFeature(feature, version)
  def createDocument(self, namespaceURI, qualifiedName, doctype):
    return self.implementation.createDocument(
      namespaceURI, qualifiedName, doctype
    )
  def createDocumentType(self, qualifiedName, publicId, systemId):
    return self.implementation.createDocumentType(
      qualifiedName, publicId, systemId
    )
  def createDOMParser(self, mode, schemaType):
    return self
  def parseURI(self, uri):
    doc= self.getDocument(uri, self)
    doc.documentURI= uri
    self.fixNode(doc.documentElement)
    return doc

  def patchModule(self):
    """ Can be extended by subclasses to perform crude run-time bug-fixing
        surgery on the implementation classes they are going to be using, the
        first time that the implementation module is loaded.
    """
    if not hasattr(self.node, 'lookupNamespaceURI'):
      self.node.lookupNamespaceURI= node_lookupNamespaceURI

  def fixNode(self, node):
    """ Can be extended by subclasses to walk the parsed DOM tree before it
        is returned, fixing up known errors in the output.
    """
    pass


class FourDOMImplementation(L2Implementation):
  """ Fixed version of PyXML's 4DOM.
  """
  def __init__(self):
    try:
      from xml.dom.DOMImplementation import implementation
      from xml.dom.ext.reader import PyExpat
      from xml.dom import FtNode, Document
    except ImportError:
      raise ImportError, (
        '4DOM is not available. If you wish to use FourDOMImplementation, '+
        'please install PyXML 0.6.6 or later.'
      )
    self.implementation= implementation
    self.node= FtNode.FtNode
    self.document= Document.Document
    self.reader= PyExpat.Reader
    L2Implementation.__init__(self)

  def patchModule(self):
    L2Implementation.patchModule(self)

    # Add XML Base attributes
    #
    from xml.dom import FtNode, Document, Element, DocumentFragment
    if not self.document._readComputedAttrs.has_key('documentURI'):
      self.document._get_documentURI= document_get_documentURI
      self.document._set_documentURI= document_set_documentURI
      self.document._readComputedAttrs['documentURI']= document_get_documentURI
      self.document._writeComputedAttrs['documentURI']= document_set_documentURI
    if not self.node._readComputedAttrs.has_key('baseURI'):
      self.node._get_baseURI= node_get_baseURI
      for subclass in (
        FtNode.FtNode, Document.Document,
        Element.Element, DocumentFragment.DocumentFragment
      ):
        subclass._readComputedAttrs['baseURI']= node_get_baseURI

    # Test for importNode wrong-ownerElement bug (PyXML <0.7).
    #
    doc= self.reader().fromString('<a b="c"/>')
    idoc= self.reader().fromString('<d/>')
    root= idoc.importNode(doc.documentElement, True)
    if root.attributes.item(0).ownerElement is not idoc:
      Element.Element._4dom_clone_= Element.Element._4dom_clone
      Element.Element._4dom_clone= fourDOM_4dom_clone

    # Test for removeAttributeNode not working (PyXML <0.7).
    #
    doc= self.reader().fromString('<a b="c"/>')
    attr= doc.documentElement.attributes.item(0)
    if attr.namespaceURI=='':
      attr.__dict__['__namespaceURI']= None
      try:
        doc.documentElement.removeAttributeNode(attr)
      except:
        Element.Element.removeAttributeNode= fourDOM_removeAttributeNode

  def getDocument(self, uri, imp):
    doc= self.reader().fromUri(uri)
    doc.__dict__['__implementation']= imp
    return doc

  def fixNode(self, node):
    """ Fix bugs in 4DOM. Various versions of 4DOM get prefix and localName
        the wrong way around in namespace declarations and set the
        namespaceURI and prefix of non-namespace attributes and default-
        namespace elements to a blank string instead of None.
    """
    d= node.__dict__
    if d['__namespaceURI']=='':
      d['__namespaceURI']= None
    if d['__prefix']=='':
      d['__prefix']= None
    if node.attributes!=None:
      for i in range(node.attributes.length):
        d= node.attributes.item(i).__dict__
        if d['__localName'] in ['', None] and d['__prefix']=='xmlns':
          d['__localName']= 'xmlns'
          d['__prefix']= None
        self.fixNode(node.attributes.item(i))
    for child in node.childNodes:
      self.fixNode(child)


class MinidomImplementation(L2Implementation):
  """ Fixed version of Python/PyXML's minidom module. Requires at least the
      version supplied with Python 2.1, as the 2.0 version is so far from
      complete even fixdom can't make it work.
  """
  def __init__(self):
    try:
      from xml.dom import minidom
    except ImportError:
      raise ImportError, (
        'Minidom is not available. If you wish to use MinidomImplementation,'+
        ' please install Python 2.1 or later, or PyXML 0.6.6 or later.'
      )
    if not hasattr(minidom, 'getDOMImplementation'):
      raise ImportError, (
        'Minidom is too old. If you wish to use MinidomImplementation, '+
        'please install Python 2.1 or later, or PyXML 0.6.6 or later.'
      )
    self.implementation= minidom.getDOMImplementation()
    self.node= minidom.Node
    self.document= minidom.Document
    L2Implementation.__init__(self)

  def patchModule(self):
    L2Implementation.patchModule(self)
    from xml.dom import minidom, pulldom
    testdoc= minidom.parseString('<el xmlns:n="u"><?pi p?></el>')

    # Add XML Base properties.
    #
    if not hasattr(testdoc, 'baseURI'):
      minidom.Node._get_baseURI= node_get_baseURI
      if not hasattr(testdoc, '__getattr__'):
        setattr(minidom.Node, 'baseURI', property(node_get_baseURI))
    if not hasattr(testdoc, 'documentURI'):
      minidom.Document._get_documentURI= document_get_documentURI
      minidom.Document._set_documentURI= document_set_documentURI
      if not hasattr(testdoc, '__getattr__'):
        setattr(minidom.Node, 'documentURI',
          property(document_get_documentURI, document_set_documentURI)
        )

    # Test for namespace-attributes-not-included-in-DOM bug (Python <2.2).
    #
    if testdoc.documentElement.attributes.length==0:
      pulldom.PullDOM._startPrefixMapping= pulldom.PullDOM.startPrefixMapping
      pulldom.PullDOM.startPrefixMapping= pulldom_startPrefixMapping
      pulldom.PullDOM._startElementNS= pulldom.PullDOM.startElementNS
      pulldom.PullDOM.startElementNS= pulldom_startElementNS

    # Test for DocumentFragment-insertion bug (Python <2.2, PyXML <0.7).
    #
    testfrag= testdoc.createDocumentFragment()
    for i in range(2):
      testfrag.appendChild(testdoc.createElementNS(None, 'el'))
    testdoc.documentElement.appendChild(testfrag)
    if len(testdoc.documentElement.childNodes)<3:
      minidom.Node._insertBefore= minidom.Node.insertBefore
      minidom.Node.insertBefore= minidom_insertBefore
      minidom.Node._appendChild= minidom.Node.appendChild
      minidom.Node.appendChild= minidom_appendChild

    # Test for lack of importNode method (Python <2.3, PyXML <0.8).
    #
    if not hasattr(testdoc, 'importNode'):
      minidom.Document.importNode= minidom_importNode
      minidom.Document.ownNodes= minidom_ownNodes

    # Test for cloneNode PI/Comment bug (PyXML 0.8.0 only).
    #
    try:
      doc.documentElement.firstChild.cloneNode(False)
    except NameError:
      for nodeType in ('PROCESSING_INSTRUCTION_NODE', 'COMMENT_NODE'):
        setattr(minidom, nodeType, getattr(minidom.Node, nodeType))

    # Test for expatreader xml: namespace bug (Python 2.2).
    #
    try:
      minidom.parseString('<el xml:lang="en"/>')
    except IndexError:
      pulldom.DOMEventStream.___init__= pulldom.DOMEventStream.__init__
      pulldom.DOMEventStream.__init__= eventstream___init__
    except KeyError:
      pass

    # Test for pulldom xml: namespace bug (Python <2.3, PyXML <0.7).
    #
    try:
      minidom.parseString('<el xml:lang="en"/>')
    except KeyError:
      pulldom.PullDOM.___init__= pulldom.PullDOM.__init__
      pulldom.PullDOM.__init__= pulldom___init__

  def getDocument(self, uri, imp):
    """ Must read content from URI manually as minidom cannot parse URIs.
    """
    from xml.dom import minidom
    resource= urllib.urlopen(uri)
    content= resource.read()
    resource.close()
    doc= minidom.parseString(content)
    doc.implementation= imp
    return doc

  def fixNode(self, node):
    """ Fix bugs in the returned DOM for older versions of minidom, which also
        use blank strings instead of None for attributes with null namespaces.
        Pyexpat also fails to detect attributes in the xml namespace on non-
        root elements.
    """
    if node.attributes!=None:
      for i in range(node.attributes.length):
        attribute= node.attributes.item(i)
        if attribute.namespaceURI=='':
          attribute.namespaceURI= None
        if attribute.prefix=='':
          attribute.prefix= None
        n= attribute.localName.split(':')
        if len(n)==2 and n[0]=='xml' and n[1]!='':
          attribute.prefix= n[0]
          attribute.localName= n[1]
          attribute.namespaceURI= XMNS
    for child in node.childNodes:
      self.fixNode(child)


# The following functions are not run directly, but are copied into the DOM
# implementation modules where they are needed.
#
def node_lookupNamespaceURI(self, prefix):
  """ Implement DOM L3 Core's prefix-to-namespaceURI lookup method. Copied
      into minidom.Node and FtNode.FtNode.
  """
  if self.nodeType==self.DOCUMENT_NODE:
    return self.documentElement.lookupNamespaceURI(prefix)
  if self.nodeType not in (self.ELEMENT_NODE, self.ATTRIBUTE_NODE):
    return None
  if prefix=='xmlns':
    return NSNS
  if prefix=='xml':
    return XMNS
  if self.nodeType==self.ELEMENT_NODE:
    if self.namespaceURI is not None and self.prefix==prefix:
      return self.namespaceURI
    xmlns= self.getAttributeNodeNS(NSNS, prefix or 'xmlns')
    if xmlns is not None:
      value= xmlns.value
      if value!='':
        return value
      return None
  if self.parentNode in (None, self.ownerDocument):
    return self.parentNode.lookupNamespaceURI(prefix)
  return None

def node_get_baseURI(self):
  """ Implement DOM L3 Core's XML Base reading method. Copied into
      minidom.Node and FtNode.FtNode.
  """
  if self.nodeType== self.ELEMENT_NODE:
    base= self.getAttributeNodeNS(XMNS, 'base')
    if base is not None:
      if urlparse.urlparse(base.value)[0]!='':
        return base.value
      if self.parentNode is None:
        return base.value
      return urlparse.urljoin(self.parentNode.baseURI, base.value)
  elif self.nodeType== self.DOCUMENT_NODE:
    return self.documentURI
  if self.parentNode is None:
    return None
  return self.parentNode.baseURI

# Implement DOM L3 Core's documentURI property. Copied into minidom.Document
# and Document.Document.
#
def document_get_documentURI(self):
  return self.__dict__.get('__fixdom_documentURI', None)
def document_set_documentURI(self, value):
  self.__dict__['__fixdom_documentURI']= value

def fourDOM_4dom_clone(self, owner):
  """ Wrap 4DOM's Element.importNode method which can get attributes wrong.
  """
  e= self._4dom_clone_(owner)
  for attr in e.attributes:
    attr._4dom_setOwnerElement(e)
  return e

def fourDOM_removeAttributeNode(self, node):
  """ Version of 4DOM's removeAttributeNode that goes wrong slightly less.
  """
  for (key, value) in self.attributes.items():
    if value is node:
      break
  else:
    import xml.dom
    raise xml.dom.NotFoundErr()
  del self.attributes[key]
  self.attributes._positions.remove(key)
  node._4dom_setOwnerElement(None)
  return node

def minidom_importNode(self, importedNode, deep):
  """ Implement the DOM 2 Core importNode method. Copied into
      xml.dom.minidom.Document.
  """
  t= importedNode.nodeType
  if t in [self.DOCUMENT_NODE, self.DOCUMENT_TYPE_NODE]:
    from xml.dom import NotSupportedErr
    raise NotSupportedErr()

  node= importedNode.cloneNode(deep or t==self.ATTRIBUTE_NODE)
  self.ownNodes(node)
  node.parentNode= None
  if t==self.ATTRIBUTE_NODE:
    node.ownerElement= None
  return node

def minidom_ownNodes(self, node):
  """ Set the ownerDocument of a node and all its child and attribute nodes.
      Recursive helper method used by importNode. Copied into
      xml.dom.minidom.Document.
  """
  node.ownerDocument= self
  if node.attributes!=None:
    for i in range(node.attributes.length):
      self.ownNodes(node.attributes.item(i))
  for child in node.childNodes:
    self.ownNodes(child)

def minidom_insertBefore(self, newChild, refChild):
  """ Implement the DOM 2 Core insertBefore method without failing on
      DocumentFragments. Copied into xml.dom.minidom.Node. Uses the old
      method stored as _insertBefore.
  """
  if newChild.nodeType==self.DOCUMENT_FRAGMENT_NODE:
    for c in newChild.childNodes[:]:
      self._insertBefore(c, refChild)
    return newChild
  else:
    return self.realInsertBefore(newChild, refChild)

def minidom_appendChild(self, newChild):
  """ Implement the DOM 2 Core appendChild method without failing on
      DocumentFragments. Copies into xml.dom.minidom.Node. Uses the old
      method stored as _appendChild.
  """
  if newChild.nodeType==self.DOCUMENT_FRAGMENT_NODE:
    for c in newChild.childNodes[:]:
      self._appendChild(c)
    return newChild
  else:
    return self._appendChild(newChild)

def pulldom___init__(self, documentFactory= None):
  """ Construct a pulldom that knows about the built-in namespaces. Copied
      into xml.dom.pulldom.PullDOM. Uses the old constructor stored as
      ___init__.
  """
  self.___init__(documentFactory)
  self._ns_contexts[-1][XMNS]= 'xml'

def pulldom_startPrefixMapping(self, prefix, uri):
  """ Store any namespace declarations in a list so they can be added back to
      the attributes list when startElementNS is called. Copied into
      xml.dom.pulldom.PullDOM. Uses old method stored as _startPrefixMapping.
  """
  if not hasattr(self, '_xmlns_attrs'):
    self._xmlns_attrs = []
  self._xmlns_attrs.append((prefix or 'xmlns', uri))
  self._startPrefixMapping(prefix, uri)

def pulldom_startElementNS(self, name, tagName , attrs):
  """ Add namespace declarations stored by startPrefixMapping to the attributes
      of the Element node. Copied into xml.dom.pulldom.PullDOM. Uses the old
      method stored as _startElementNS.
  """
  self._startElementNS(name, tagName, attrs)
  nss= getattr(self, '_xmlns_attrs', [])
  if len(nss)>0:
    if hasattr(self, 'pop'):
      node= self.pop()
    else:
      node= self.curNode
    for prefix, uri in nss:
      if prefix!='xmlns':
        prefix= 'xmlns:'+prefix
      attr= self.document.createAttributeNS(NSNS, prefix)
      attr.value= uri
      node.setAttributeNodeNS(attr)
    if hasattr(self, 'pop'):
      self.push(node)
    self._xmlns_attrs[:]= []

def eventstream___init__(self, stream, parser, bufsize):
  """ Construct a DOMEventStream after checking if the parser supplied is a
      broken xml.sax.expatreader that blows up when presented with an xml:
      namespace. If it is, hack its namespace stack so it doesn't. Copied into
      xml.dom.pulldom.DOMEventStream. Uses the old constructor stored as
      ___init__.
  """
  if hasattr(parser, '_ns_stack') and parser._ns_stack==[]:
    parser._ns_stack.append({XMNS: ['xml']})
  self.___init__(stream, parser, bufsize)
